//`define bbug 1
`define fastflash 1
`timescale 1ns / 1fs
`default_nettype none

/* for //e		*/

module C302 		// this version is for the //e  
						// 6 bit code & 12bit RGB 
						// has SHR added for //e use
  (					// it has code added to flash the M25P80 from the Apple
						// also has the h_sync & v_sync changed to be more friendly with HDTVs
						// same as C300 except changes to improve SHR output
						
   input 	BUS_3M58,		// on a IIgs this is M2B0
   input	 	BUS_7M,
   inout [15:0]  BUS_A,
   inout [7:0]  BUS_D, 
	output 	BUS_DMA, 
   input 	BUS_DMAIN,
   output   BUS_DMAOUT,
	output   BUS_INH,
   input 	BUS_INTIN,
   output   BUS_INTOUT,
   input    BUS_IOSTROBE,
   input    BUS_IRQ,
   input    BUS_NMI,
   input    BUS_PHASE0,
   input    BUS_PHASE1,
   input    BUS_Q3,
   input    BUS_R_W,
   output   BUS_RDY,
   input    BUS_RESET,
   input    BUS_SYNC,
   input    BUS_USER1,

   output   IDE_CS1,
   output [13:0]   IDE_D,
   input   IDE_RES,
   output   IDE_ACT,   
   input    N_BUS_DEV_SEL,
   input    N_BUS_IO_SEL,
   output [16:0]   SRAM_A,
   inout  [7:0]    SRAM_D,
   output   SRAM_NCS0,
   output   SRAM_NCS1,

   output   SRAM_NOE,
   output   SRAM_NWE,
	
   input 	SYSCLK,
   output [5:0]  RGB,
   output   v_sync,
   output   h_sync,
   input   	SD_DET,
   input    SOFT_TCK,
   input    SOFT_TDI,
   input    SOFT_TMS,
   output   SOFT_TDO,

   input    spi_sdo, 
   output   spi_sdi, 
   output   spi_sck, 
   output   reg spi_rom_cs	
  );
/* version */
	reg [7:0] maj_version = 03;  // use 255 for test versions
	reg [7:0] min_version = 02;
/* Apple clocks sync'd to clk50 */
	reg reset_s;	 
	reg phase0_s;	
	reg Q3_s;		
	reg devsel_s;		
	reg iosel_s;	
	reg iostrobe_s;			
//	reg m2b0_s;		
	reg[15:0] bus_a_1;	
	reg[15:0] bus_a_s; 	
	reg[7:0] bus_d_1;		
	reg[7:0] bus_d_s;
//	reg user1_s;
	reg ph1_s;		
	/* ---------------------------- */

	reg clk25_b;

	wire io_read;
	wire [7:0] io_reg;
	wire sck;
	reg mosi = 0;
	wire clk25;
	wire bugin;		// used for debugging
	wire 	jumper = IDE_RES & SOFT_TCK; // low if either jumper in place - changed to one & 5/14/11
//	wire jumper = 1'b1;  // fake jumper off
   wire [7:0] 	write_data;
   wire [15:0] video_wr_addr;  
`ifdef fastflash
	reg[8:0] ff_addr = 9'h000;		// fast flash uses hgr2 addr 
	reg[8:0] fr_addr = 9'h000;		// read back & error check
`endif
   wire 	TxD;

   wire 	video_read;
   wire 	video_we;
   wire 	video_write_SRAM;
	wire pal_we;
	
   assign	IDE_CS1 = SD_DET;

   wire 	CLKFX_OUT;
   wire 	CLKIN_IBUFG_OUT,CLK0_OUT,LOCKED_OUT,LOCKED_OUT_150,CLK0_OUT_150;
   wire 	clk50 = CLKFX_OUT;

	/* --- Apple IIgs must use mask when asserting INH --- */
	/* 	 We don't use INH in this version					 */

	assign 	BUS_INH = 1'bz;	
   assign 	SRAM_NCS0 = 0;
   assign 	SRAM_NCS1 = 1;
   assign 	BUS_DMAOUT = BUS_DMAIN;		// just connect the DMA chain 
	assign 	BUS_INTOUT = BUS_INTIN;		// just connect the interrupt chain
	assign	BUS_DMA = 1'bz;				// we don't use DMA	
   assign 	SOFT_TDO = TxD;
	
   wire [10:0] rom_addr;
   wire [7:0] 	char_data; 		// from font
   wire [7:0] 	video_ram_data;

   wire [16:0] ram_address;	// address to read from video buffer 
   wire 	update_time;
   wire 	auxmem; 	// if 1 means the apple wrote a video address in the bank 1 (aux memory)
	
	/* --- video_wr_addr (16 bit) is the ram address output from the write_clk module --- */
	/* --- on the falling edge of Q3 and rw == 0	 and (user1_s == 0)						 --- */
	/* --- ram_address is the address from the vga_video module to read data from the --- */
	/* --- SRAM or BRAM (on FPGA block RAM)							  							 --- */
	wire write_vid_is_naux = (video_wr_addr > 16'h1FFF) & (video_wr_addr < 16'h6000) & ~auxmem;
	wire read_vid_is_naux = (ram_address[15:0] > 16'h1FFF) & (ram_address[15:0] < 16'h6000) & (ram_address[16] == 1'b0);
	wire[13:0] BRAM_AA_NAUX;
	wire[13:0] BRAM_AB_NAUX;
	wire[7:0] BRAM_DA_NAUX;
	wire[7:0] BRAM_DB_NAUX;

	wire[11:0] BRAM_AA;	// write addr
	wire[11:0] BRAM_AB;	// read addr
	wire[7:0] BRAM_DA;	// write data
	wire[7:0] BRAM_DB;	// read data
	
	wire[7:0] BRAM_AA_PAL;	// write addr
	wire[7:0] BRAM_AB_PAL;	// read addr
	wire[11:0] BRAM_DA_PAL;	// write data
	wire[11:0] BRAM_DB_PAL;	// read data
	
	wire[7:0] pal_addr; // read address
	wire[11:0] pal_data; // read data	

	wire write_vid_is_text = (video_wr_addr > 16'h03FF) & (video_wr_addr < 16'h0C00);
	wire read_vid_is_text = (ram_address[15:0] > 16'h03FF) & (ram_address[15:0] < 16'h0C00);
	wire video_mem = ((video_wr_addr > 16'h1FFF) && (video_wr_addr < 16'h9D00)) ? 1'b1 : 1'b0;	
	wire vid_read_mem = ((ram_address[15:0] > 16'h1FFF) && (ram_address[15:0] < 16'h9D00)) ? 1'b1 : 1'b0;
	wire write_vid_is_gs = ((video_wr_addr > 16'h9CFF) && (video_wr_addr < 16'hA000) && auxmem) ? 1'b1 : 1'b0;
`ifdef fastflash
	wire write_vid_is_ff = ((video_wr_addr > 16'h3FFF) && (video_wr_addr < 16'h4100) && ~auxmem && video_write_SRAM);
`endif	
	wire write_is_scb = ((video_wr_addr > 16'h9CFF) && (video_wr_addr < 16'h9DC8) && auxmem) ? 1'b1 : 1'b0;
	wire write_is_even = (video_wr_addr[0] == 1'b0) ? 1'b1 : 1'b0; 
	wire write_is_pal =  ((video_wr_addr > 16'h9DFF) && (video_wr_addr < 16'hA000) && auxmem ) ? 1'b1 : 1'b0;

	assign 	SRAM_A = (video_write_SRAM && ~write_vid_is_text && ~write_vid_is_naux  && ~write_vid_is_gs) ? {auxmem, video_wr_addr} : (vid_read_mem  && ~read_vid_is_naux) ? ram_address : 17'bz;
	assign 	SRAM_D = (video_write_SRAM & video_mem & ~write_vid_is_naux && ~write_vid_is_gs) ? write_data : 8'bz;	

	assign 	video_ram_data = (~video_write_SRAM & vid_read_mem & ~read_vid_is_naux) ?  SRAM_D : read_vid_is_text ? BRAM_DB : (~video_write_SRAM & read_vid_is_naux) ? BRAM_DB_NAUX : 8'bz;	

	assign 	BRAM_AA = (video_write_SRAM && write_vid_is_text) ? {auxmem, video_wr_addr[10:0]} : 12'bz; 	// write addr
	assign 	BRAM_AB = (read_vid_is_text) ? {ram_address[16], ram_address[10:0]} : 12'bz;						// read addr
	assign	BRAM_DA = (video_write_SRAM && write_vid_is_text) ? write_data : 8'bz;

	assign 	BRAM_AA_NAUX = (video_write_SRAM && write_vid_is_naux) ? video_wr_addr[13:0] : 14'bz; 	// write
	assign 	BRAM_AB_NAUX = (read_vid_is_naux) ? ram_address[13:0] : 14'bz;						// read addr	
	assign	BRAM_DA_NAUX = (video_write_SRAM && write_vid_is_naux) ? write_data : 8'bz;

reg [7:0] even_pal = 0;
reg [3:0] odd_pal = 0;

	assign	BRAM_AA_PAL = (video_write_SRAM && write_is_pal) ? (video_wr_addr[8:1]) : 8'bz;	
	assign	BRAM_DA_PAL = (video_write_SRAM && write_is_pal) ? {odd_pal, even_pal} : 12'bz;

	assign	BRAM_AB_PAL = pal_addr;
	assign 	pal_data = BRAM_DB_PAL;
	
	assign 	SRAM_NOE = ~video_read;
   assign 	SRAM_NWE = ~video_we;
/* -------------------- scan control bytes ----------------------------------------------	*/	
wire[7:0] scb_addr; // read address
wire[7:0] scb_data; // read data
reg [7:0] scb [199:0];
`ifdef fastflash
reg [7:0] ff_array [255:0];
`endif

	always @(posedge clk50)
		begin
		if (video_write_SRAM && write_is_scb) scb [video_wr_addr[7:0]] <= write_data; 
		if(video_write_SRAM && write_is_pal && write_is_even) even_pal <= write_data;
		if(video_write_SRAM && write_is_pal && ~write_is_even) odd_pal <= write_data[3:0];
`ifdef fastflash
		if(write_vid_is_ff) ff_array[video_wr_addr[7:0]] <= write_data;
`endif
		end
		
reg [255:0] k;	// was reg [199:0]

	initial
		begin
		for (k = 0; k < 200; k = k + 1)
			begin
			scb[k] = 0;
			end
`ifdef fastflash
		for (k = 0; k < 256; k = k + 1)
			begin 
			ff_array [k] = 0;
			end
`endif			
		end

	assign scb_data = scb[scb_addr];

text_buf text_buf (	// buffers text pages 1 & 2, main and aux
	.clka(clk50),
	.wea(video_we),   // enable writing
	.addra(BRAM_AA), 	// write address 12 bit
	.dina(BRAM_DA),  	// write data
	.clkb(clk50),
	.addrb(BRAM_AB), 	// read address 12 bit 
	.doutb(BRAM_DB)); // read data

aux_buf aux_buf(				// buffers main RAM HGR & DHR both pages
	.clka(clk50),
	.wea(video_we),  			// enable writing
	.addra(BRAM_AA_NAUX), 	// write address 14 bit
	.dina(BRAM_DA_NAUX), 	// write data 
	.clkb(clk50),
	.addrb(BRAM_AB_NAUX), 	// read address 14 bit 
	.doutb(BRAM_DB_NAUX)); 	// read data 
	
pal_buf pal_buf (				// buffers palettes 12 bits
									// not used if pal_fixed defined in video module
	.clka(clk50),
	.wea(video_we), 			// enable writing
	.addra(BRAM_AA_PAL), 	// write address 8 bit
	.dina(BRAM_DA_PAL), 		// write data 12 bit
	.clkb(clk50),
	.addrb(BRAM_AB_PAL), 	// read address 8 bit 
	.doutb(BRAM_DB_PAL)); 	// read data 12 bit 

	/* --------------------------------------------------------------------------------- */

	wire reset = ~reset_s | ~LOCKED_OUT;	
	assign 	BUS_RDY = 1'bz;
   wire [5:0]  RGB_from_module;						 
	assign RGB = RGB_from_module;
//	wire [15:0] debug;
	reg [7:0] txt_bk = 8'hF0;	// holds text text color hi 4bits - bkgnd color lo 4bits
	reg [3:0] bdr = 4'h0;		// holds border color
	reg bw_dhr = 0;				// black & white dhr
//	reg opt12 = 1;					// optimize for 12 bit color // no longer used 5/19/13
	reg init_done = 0;
	
   vga_video 
     video_module (
		   .clk50(clk50), 
		   .char_data(char_data), 
		   .ram_data(video_ram_data), 
		   .bus_addr(bus_a_s), 
			.bus_data(bus_d_s),	
		   .ph1_s(ph1_s),			
			.phase0_s(phase0_s), 
			.Q3(Q3_s),				
		   .rw(BUS_R_W), 
		   .reset_n(~reset), 
			.iosel_s(iosel_s), 
			.iostrobe_s(iostrobe_s),
			.devsel_s(devsel_s),					
			.BUS_SYNC(BUS_SYNC), // only on slot 7 - added to sync CB video with apple video  
`ifdef bbug			
			.bugin(bugin),			// added 10/1/10 to pass signal for bug border
`endif
	//		.debug(debug),			// added 1/24/12
			.init_done(init_done),		
			.txt_bk(txt_bk),		
			.bdr(bdr),			
			.bw_dhr(bw_dhr),
	//		.opt12(opt12),		
			.scb_data(scb_data),			// input scanline from BRAMwith
			.scb_addr(scb_addr),			// output scanline address in BRAM
			.pal_data(pal_data),		// input palette even byte from BRAM
			.pal_addr(pal_addr),		// output palette even byte address in BRAM
			
		   .update_time(update_time),
			.auxmem(auxmem),
		   .video_read(video_read),
		   .v_sync(v_sync), 
		   .h_sync(h_sync), 
		   .rom_addr(rom_addr), 
		   .ram_address(ram_address),
			.IDE_D(IDE_D),				// added 12/17/11 for RGB 12 bit
			.clk25(clk25),
			.RGB(RGB_from_module),
			.io_read(io_read), // output added 4/16/13 - 1 when reading i/o location
			.io_reg(io_reg)   // output added 4/16/13 - contains data to put on io bus
		   );

   write_clk 
     write_clk (
		.clk(clk50), 
		.ph0(BUS_PHASE0), 	// not replaced with phase0_s since sync'd in write_clk 9/25/10
		.Q3(BUS_Q3), 			// sync'd in write_clk 9/25/10
		.rw(BUS_R_W),
		.update_time(update_time), 
		.bus_a_s(bus_a_s), 
		.bus_d_s(bus_d_s),
//		.debug(debug),			// added 1/25/12
		.we(video_we),
		.video_write_SRAM(video_write_SRAM),
		.ram_addr(video_wr_addr), 
		.write_data(write_data)
		);

   clk0 clock (
	       .CLKIN_IN(SYSCLK), 
		//	 .CLKIN_IN(BUS_7M),
	       .RST_IN(1'b0 /*~BUS_RESET*/),
	       .CLKFX_OUT(CLKFX_OUT), 
	       .CLKIN_IBUFG_OUT(CLKIN_IBUFG_OUT), 
	       .CLK0_OUT(CLK0_OUT), 
	       .LOCKED_OUT(LOCKED_OUT)
	       );

   font font0 (
	       .addr(rom_addr),
	       .dout(char_data),
	       .clk(clk50)
	       );



/* ---------------- spi ----------------	*/
	assign spi_sdi = mosi;	// to spi_sdi (serial data into device)
   assign spi_sck = sck;	// to spi_sck (clock for device)

	wire [7:0] spi_out;
	wire [7:0] spi_in;
	wire write_spi;
	wire spi_done;
	reg fft = 0;
// The following spi code is taken from the spi.v module.
// See copyright and license below:
// ====================================================================
//                         VECTOR-06C FPGA REPLICA
//
// 		Copyright (C) 2007, Viacheslav Slavinsky
//
// This core is distributed under modified BSD license. 
// For complete licensing information see LICENSE.TXT.
// --------------------------------------------------------------------  
   wire	miso = spi_sdo;		// from spi_sdo (serial data from device)
   wire [7:0] di = spi_in;		// 8 bit parallel data in
   wire	wr = write_spi;		// 1 = write to spi
   wire [7:0] do;					// 8 bit parallel data out
	assign spi_out = do;
   reg	done;						// 1 = 8 bits shifted in/out
	assign spi_done = done;
   reg [7:0] shiftreg;
   reg [7:0] shiftski;
   reg [1:0] state = 0;
   reg scken = 0;

   assign sck = clk25_b & scken;	// scken used to switch on/off device clock
   assign do  = shiftreg;

   always @(posedge clk50 or negedge reset_s) 
		begin
      if (!reset_s) 
			begin
			state <= 0;			// reset state
			mosi <= 1'b0;		
			done  <= 0;			// not done
			scken <= 0;			// device clock off
			end 
		else 
			begin
			if (clk25_b)		// every other time 
				begin
				case (state)
				0:	begin
					if (wr) 
						begin
						done <= 1'b0;			
						state <= 1;					
						shiftreg <= di;				// load the data
						shiftski <= 8'b11111111;	// use as counter
						end
					end
				1: begin
					scken <= 1;							// device clock on
					mosi <= shiftreg[7];				// send a bit to device
					shiftreg <= {shiftreg[6:0],miso};	// left shift data & shift bit in from device
					shiftski <= {1'b0,shiftski[7:1]};	// right shift counter
					if (|shiftski == 0)	
						begin 							// counter is 0
						state <= 2;					
						scken <= 0;						// device clock off
						end
					end
				2:	begin
		//			mosi <= 1'b0; // shouldn't be necessary but a nice debug view (VS)
					done <= 1'b1;
					state <= 0;
					end
				default: ;
				endcase
			end
      end
   end
/* -------------------- end spi ------------	*/
/* ---- test peripheral I/O ----------------		*/
	reg [7:0] C0n0 = 0;
	reg [7:0] C0n1 = 0;
	reg [7:0] C0n2 = 0;
	reg [7:0] C0n3 = 0;
	reg [7:0] C0n4 = 0;
	reg [7:0] C0n5 = 0;
	reg [7:0] C0n6 = 0;
	reg [7:0] C0n7 = 0;
	reg [7:0] C0nC = 0;
	reg [7:0] C0nF = 8'hCB;	// ID byte read only
	reg [7:0] bus_reg = 0;
	reg rw_s = 0;
	wire device_io = (devsel_s == 1'b0) && (ph1_s == 1'b0); // && (user1_s == 1'b0);
	wire dev_read = device_io & rw_s;
	wire dev_write = device_io & ~rw_s;
//	assign debug = {C0n1,C0n0};
	reg write_spi_reg = 0;
	reg write_spi_reg4 = 0;	// fastflash
	reg write_spi_reg6 = 0;	// read back error check
	reg [7:0] spi_in_reg = 0;
	reg [7:0] spi_in_reg4 = 0;
	reg [7:0] spi_in_reg6 = 0;
	wire one_byte_cmd;
	wire two_byte_cmd;
	reg write_cmd_reg = 0;
	reg [7:0] cmd_reg = 0;
`ifdef fastflash
	reg ff_on = 0;				// fastflash on/off
	reg fr_on = 0;				// fast read on/off
`endif
	wire read_cmds;
`ifdef fastflash
	assign write_spi = (one_byte_cmd || two_byte_cmd) ? write_cmd_reg : (ff_on) ? write_spi_reg4 : fr_on ? write_spi_reg6 : write_spi_reg;
	assign spi_in = (one_byte_cmd || two_byte_cmd) ? cmd_reg : ff_on ? spi_in_reg4 : fr_on ? spi_in_reg6 : spi_in_reg;
`else
	assign write_spi = (one_byte_cmd || two_byte_cmd) ? write_cmd_reg : write_spi_reg;
	assign spi_in = (one_byte_cmd || two_byte_cmd) ? cmd_reg : spi_in_reg;

`endif
	assign read_cmds = (C0n1 == 8'h03) || (C0n1 == 8'h05);
	reg do_nvram = 0;

	reg cmd_cs = 1;
	reg spi_cs = 1;
	reg spi_cs4 = 1;			// fastflash page program
	reg spi_cs6 = 1;			// read back error check
	reg [7:0] page_errs = 0;
	reg [4:0] nvram_state = 0;
	reg [4:0] nvram_state4 = 0;  // fastflash 
	reg [4:0] nvram_state6 = 0;  // read back error check 
	
	always @(posedge clk50)
		begin
		clk25_b <= ~clk25_b;
`ifdef fastflash
		if (clk25_b) spi_rom_cs <= (one_byte_cmd || two_byte_cmd) ? cmd_cs : ff_on ? spi_cs4 : fr_on ? spi_cs6 : spi_cs;
`else
		if (clk25_b) spi_rom_cs <= (one_byte_cmd || two_byte_cmd) ? cmd_cs : spi_cs; 
`endif
		end

	always @(negedge Q3_s)	
		begin
			if (do_nvram == 1'b1) do_nvram <= 1'b0;
			if((spi_rom_cs == 1'b1)) 
				begin
				do_nvram <= 1'b0;
				if ((C0n0 == 8'h02) && read_cmds) C0n2 <= spi_out;
				if ((C0n0 == 8'h06))
					begin
						C0n7 <= page_errs;
						C0n2 <= nvram_state6;	
					end					
				end
		if(dev_write)
		 	begin
			case(bus_a_s[3:0])
			4'h0:							// commands 0 = reset regs, 1 = version, 2 = M25P80, 5 = test/debug
											// 3 - do nothing, 4 - fastflash, 6 - error test
				begin
				C0n0 <= bus_d_s;  	// make it stick
				if(bus_d_s == 8'h00)	// reset io regs command
					begin
					C0n1 <= 8'h00;
	//				C0n2 <= 8'h00;
					C0n3 <= 8'h00;
					C0n4 <= 8'h00;
					C0n5 <= 8'h00;
					C0n6 <= 8'h00;
					C0n7 <= 8'h00;
					C0nC <= 8'h00;
					end
				if(bus_d_s == 8'h01)		// version
					begin
					C0n1 <= min_version;
					C0n2 <= maj_version;
					end					
				end

			4'h1: 
				begin
				C0n1 <= bus_d_s;		// data low byte
				end
				
			4'h2: C0n2 <= bus_d_s;
			4'h3:	begin
					C0n3 <= bus_d_s;
`ifdef fastflash
					if((C0n0 == 8'h02) || (C0n0 == 8'h04) || (C0n0 == 8'h06))
`else
					if(C0n0 == 8'h02)
`endif					
						begin
						do_nvram <= 1'b1;
						end
					end			
			4'h4: C0n4 <= bus_d_s;		// address low		
			4'h5: C0n5 <= bus_d_s;		// address mid
			4'h6: C0n6 <= bus_d_s;		// address high
			4'h7: C0n7 <= bus_d_s;		// page errors (fastflash)
			4'hC: C0nC <= bus_d_s;			

			endcase
			end
		if(dev_read)
			begin
			case(bus_a_s[3:0])
			4'h0: bus_reg <= C0n0; 
			4'h1: bus_reg <= C0n1; 
			4'h2: bus_reg <= C0n2; 
			4'h3: begin
					bus_reg <= C0n3;
`ifdef fastflash					
					if((C0n0 == 8'h02) || (C0n0 == 8'h04) || (C0n0 == 8'h06)) do_nvram <= 1'b1;
`else
					if(C0n0 == 8'h02) do_nvram <= 1'b1;
`endif
					end
			4'h4: bus_reg <= C0n4; 
			4'h5:	bus_reg <= C0n5; 
			4'h6: bus_reg <= C0n6;
			4'h7: bus_reg <= C0n7;
			4'hC: bus_reg <= C0nC;
			4'hF:	bus_reg <= C0nF;
			endcase
			end
		end
	assign BUS_D = dev_read ? bus_reg : io_read ? io_reg : 8'bz;	
/* ---- initial set-up from flash RAM data (run once) ----	*/
/*      use with read routine below									*/
reg [3:0] init_state = 0; 
reg init_set = 1;				// routine runs only once at startup
reg [7:0] init_cnt = 8'h0; // so we can get up to 256 bytes

always @(posedge clk25_b)
	begin
	if(init_set)
		begin
			if((init_state == 13) && (spi_done))
				begin
				case (init_cnt)
				8'h03: if (spi_out != 8'hC3) init_set <= 1'b0;	// $3 - $6 test for magic #
				8'h04: if (spi_out != 8'hC2) init_set <= 1'b0;
				8'h05: if (spi_out != 8'h53) init_set <= 1'b0;
				8'h06: if (spi_out != 8'h30) init_set <= 1'b0;
				8'h10: txt_bk <= spi_out;
				8'h11: bdr <= spi_out[3:0];
				8'h12: bw_dhr <= spi_out[5];
		//		8'h16: opt12 <= spi_out[0];  // no longer used 5/19/13
				8'hFF: begin
						 init_set <= 1'b0;
						 init_done <= 1'b1;
						 end
				default:;
				endcase
		//		init_cnt <= init_cnt + 1;
				end
		end
	end
	
/* ---- read, program or sector erase NVRAM (M25) ------------------		*/
/* Read Data Bytes (READ)	$03
	Page Program	 (PP)		$02
	Sector Erase	 (SE)		$D8
*/
always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)
		begin
		spi_cs <= 1'b1;
		write_spi_reg <= 1'b0;
		nvram_state <= 0;
		end
	else if (init_set)
		begin
		case (init_state)
		0: begin
			init_state <= 1;
			spi_cs <= 1'b0;					// turn on M25
			spi_in_reg <= 8'h03;				// read
			write_spi_reg <= 1'b1;			// write command seq		
			end
		1: init_state <= 2;	
		2: begin										
			write_spi_reg <= 1'b0;				
			if (spi_done) 
				begin
				init_state <= 3;
				end
			end
		3: begin										
			spi_in_reg <= 8'h08;					// write high addr byte
			init_state <= 4;
			write_spi_reg <= 1'b1;
			end
		4: init_state <= 5;
		5: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				init_state <= 6;
				end
			end
		6: begin									
			spi_in_reg <= 8'h00;					// write mid addr byte
			init_state <= 7;
			write_spi_reg <= 1'b1;
			end
		7: init_state <= 8;
		8: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				init_state <= 9;
				end
			end
		9: begin							
			spi_in_reg <= init_cnt;				// write low addr byte
			init_state <= 10;		
			write_spi_reg <= 1'b1;
			end
		10: init_state <= 11;
		11:begin									
			write_spi_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin
				spi_in_reg <= 8'h00;
				init_state <= 12;
				write_spi_reg <= 1'b1;
				end
			end
		12: init_state <= 13;
		13:begin
			write_spi_reg <= 1'b0;		
			if (spi_done)
				begin
				spi_cs <= 1'b1;				// turn off M25
				init_state <= 14;
				end
			end
		14:begin
			init_cnt <= init_cnt + 1;
			init_state <= 0;
			end
		default: ;
		endcase
		end
	/* read - sector erase - slow flash - one byte at a time	*/
	else if(((C0n1 == 8'h03) || (C0n1 == 8'h02) || (C0n1 == 8'hD8)) && (C0n0 == 8'h02))    
		begin
		case (nvram_state)
		0: begin										
			if(do_nvram) 
				begin
				nvram_state <= 1;
				spi_cs <= 1'b0;					// turn on M25
				spi_in_reg <= C0n1;				// read, program or sector erase
				write_spi_reg <= 1'b1;			// write command seq
				end
			end
		1: nvram_state <= 2;	
		2: begin										
			write_spi_reg <= 1'b0;				
			if (spi_done) 
				begin
				nvram_state <= 3;
				end
			end
		3: begin										
			spi_in_reg <= C0n6;					// write high addr byte
			nvram_state <= 4;
			write_spi_reg <= 1'b1;
			end
		4: nvram_state <= 5;
		5: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				nvram_state <= 6;
				end
			end
		6: begin									
			spi_in_reg <= C0n5;					// write mid addr byte
			nvram_state <= 7;
			write_spi_reg <= 1'b1;
			end
		7: nvram_state <= 8;
		8: begin										
			write_spi_reg <= 1'b0;
			if (spi_done) 
				begin
				nvram_state <= 9;
				end
			end
		9: begin							
			spi_in_reg <= C0n4;				// write low addr byte
			nvram_state <= 10;		
			write_spi_reg <= 1'b1;
			end
		10: nvram_state <= 11;
		11:begin									
			write_spi_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin
				if (C0n1 == 8'hD8)			// if sector erase
					begin
					spi_cs <= 1'b1;			// turn off M25
					nvram_state <= 0;					
					end
				else
					begin
					spi_in_reg <= (C0n1 == 8'h02) ? C0n2 : 8'h00;
					nvram_state <= 12;
					write_spi_reg <= 1'b1;
					end
				end
			end
		12: nvram_state <= 13;
		13:begin
			write_spi_reg <= 1'b0;		
			if (spi_done)
				begin
				spi_cs <= 1'b1;				// turn off M25
				nvram_state <= 0;
				end
			end
		default: ;
		endcase
		end
	end
	
`ifdef fastflash
	/* Fast Flash 	*/
	/* uses page program to flash 256 bytes from SRAM */
always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)
		begin
		spi_cs4 <= 1'b1;
		write_spi_reg4 <= 1'b0;
		nvram_state4 <= 0;
		ff_addr <= 9'h000;
		ff_on  <= 1'b0;
		end
	else if  ((C0n1 == 8'h02) && (C0n0 == 8'h04))		// fast flash using page program
		begin
		case (nvram_state4)
		0: begin										
			if(do_nvram) 
				begin
				ff_on <= 1'b1;
				nvram_state4 <= 1;
				spi_cs4 <= 1'b0;					// turn on M25
				spi_in_reg4 <= C0n1;				// program only 
				write_spi_reg4 <= 1'b1;			// write command seq
				end
			end
		1: nvram_state4 <= 2;	
		2: begin										
			write_spi_reg4 <= 1'b0;				
			if (spi_done) 
				begin
				nvram_state4 <= 3;
				end
			end
		3: begin										
			spi_in_reg4 <= C0n6;					// write high addr byte
			nvram_state4 <= 4;
			write_spi_reg4 <= 1'b1;
			end
		4: nvram_state4 <= 5;
		5: begin										
			write_spi_reg4 <= 1'b0;
			if (spi_done) 
				begin
				nvram_state4 <= 6;
				end
			end
		6: begin									
			spi_in_reg4 <= C0n5;					// write mid addr byte
			nvram_state4 <= 7;
			write_spi_reg4 <= 1'b1;
			end
		7: nvram_state4 <= 8;
		8: begin										
			write_spi_reg4 <= 1'b0;
			if (spi_done) 
				begin
				nvram_state4 <= 9;
				end
			end
		9: begin							
			spi_in_reg4 <= 8'h00;				// write low addr byte (always zero)
			nvram_state4 <= 10;		
			write_spi_reg4 <= 1'b1;
			ff_addr <= 9'h000;
			end
		10: begin
			 nvram_state4 <= 11;
			 end
		11:begin	
			write_spi_reg4 <= 1'b0;			// stop writing
			if (spi_done)
				begin								// this is where we flash 256 bytes 
				spi_in_reg4 <= ff_array[ff_addr[7:0]];
				ff_addr <= ff_addr + 1;
				if (ff_addr == 9'h100) 
					begin
					nvram_state4 <= 12;
					ff_addr <= 9'h000;  		// reset		
					end
				else nvram_state4 <= 10;
				write_spi_reg4 <= 1'b1;
				end
			end
		12:begin
			nvram_state4 <= 13;			
			end
		13:begin
			write_spi_reg4 <= 1'b0;		
			if (spi_done)						
				begin			
				spi_cs4 <= 1'b1;				// turn off M25
				nvram_state4 <= 0;
				ff_on  <= 1'b0;				// fastflash is off
				end
			end
		default: ;
		endcase
		end
	end	
/* read back & test for errors */
always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)
		begin
		spi_cs6 <= 1'b1;
		write_spi_reg6 <= 1'b0;
		nvram_state6 <= 0;
		fr_addr <= 9'h000;
		fr_on  <= 1'b0;
		page_errs <= 8'h00;
		end
	else if  ((C0n1 == 8'h03) && (C0n0 == 8'h06))
		begin
		case (nvram_state6)
		0:	begin
			if(do_nvram) 
				begin
				fr_on <= 1'b1;
				fr_addr <= 9'h000;
				nvram_state6 <= 1;
				spi_in_reg6 <= 8'h05;	// read status 
				write_spi_reg6 <= 1'b1;
				spi_cs6 <= 1'b0;
				end
			end
		1: nvram_state6 <= 2;
		2: begin										
			write_spi_reg6 <= 1'b0;				
			if (spi_done) nvram_state6 <= 3;
			end
		3: begin
			spi_in_reg6 <= 8'h00;		// dummy byte
			nvram_state6 <= 4;
			write_spi_reg6 <= 1'b1;		// write command seq
			end
		4: nvram_state6 <= 5;
		5: begin										
			write_spi_reg6 <= 1'b0;				
			if (spi_done) 
				begin
				if(spi_out[0] == 1'b0) 
					begin
					spi_cs6 <= 1'b1;					
					nvram_state6 <= 6;
					page_errs <= 8'h00;
					end
				else 
					begin
					nvram_state <= 1;
					spi_in_reg6 <= 8'h05;	// read status 
					write_spi_reg6 <= 1'b1;					
					end
				end
			end
		6: begin
			spi_in_reg6 <= C0n1;			// read
			nvram_state6 <= 7;
			spi_cs6 <= 1'b0;
			write_spi_reg6 <= 1'b1;		// write command seq
			end
		7: nvram_state6 <= 8;
		8: begin										
			write_spi_reg6 <= 1'b0;				
			if (spi_done) nvram_state6 <= 9;
			end
		 9:begin										
			spi_in_reg6 <= C0n6;					// write high addr byte
			nvram_state6 <= 10;
			write_spi_reg6 <= 1'b1;
			end
		10: nvram_state6 <= 11;
		11:begin										
			write_spi_reg6 <= 1'b0;
			if (spi_done) nvram_state6 <= 12;
			end
		12:begin									
			spi_in_reg6 <= C0n5;					// write mid addr byte
			nvram_state6 <= 13;
			write_spi_reg6 <= 1'b1;
			end
		13:nvram_state6 <= 14;
		14:begin										
			write_spi_reg6 <= 1'b0;
			if (spi_done) nvram_state6 <= 15;
			end
		15:begin							
			spi_in_reg6 <= fr_addr[7:0];			// write low addr byte
			nvram_state6 <= 16;			
			write_spi_reg6 <= 1'b1;
			end
		16:nvram_state6 <= 17;
		17:begin
			write_spi_reg6 <= 1'b0;		
			if (spi_done)
				begin
				nvram_state6 <= 18;			
				write_spi_reg6 <= 1'b1;				
				end
			 end
		18:nvram_state6 <= 19;
		19:begin	
			write_spi_reg6 <= 1'b0;		
			if (spi_done)
				begin
				if(spi_out != ff_array[fr_addr[7:0]]) page_errs <= page_errs + 1;
				fr_addr <= fr_addr + 1;
				if(fr_addr == 9'h100) 
					begin
					fr_addr <= 9'h000;
					nvram_state6 <= 20;
					end
				else 
					begin
					nvram_state6 <= 6; //17; 		
					spi_cs6 <= 1'b1;
					end
				end
			end
		20:begin
				begin
				spi_cs6 <= 1'b1;				// turn off M25
				nvram_state6 <= 0;
				fr_on <= 1'b0;
				end
			end
		default: ;
		endcase
		end
	end
`endif
/* --- one byte NVRAM commands ---	*/
/* Write enable (WREN) 	$06
	Write disable (WRDI)	$04
	Bulk Erase (BE)		$C7
	Deep Power Down (DP)	$B9
	Release Deep Power Down (RES)	$AB
*/
/* --- two byte NVRAM commands ---	*/
/* Read Status Register (RDSR)	$05
	Write Status Register (WRSR)	$01
*/
reg [2:0] cmd_state = 0;

assign one_byte_cmd = (C0n1 == 8'h06) || (C0n1 == 8'h04) || (C0n1 == 8'hC7) || (C0n1 == 8'hB9) || (C0n1 == 8'hAB) && (C0n0 == 8'h02);
assign two_byte_cmd = (C0n1 == 8'h05) || (C0n1 == 8'h01) && (C0n0 == 8'h02);
always @(posedge clk25_b or negedge reset_s)
	begin
	if(reset_s == 1'b0)
		begin
		cmd_cs <= 1'b1;
		cmd_state <= 0;
		write_cmd_reg <= 1'b0;
		end
	else if((C0n0 == 8'h02) && (one_byte_cmd || two_byte_cmd))
		begin
		case (cmd_state)
		0: begin									// use 1 cycle
			if (do_nvram)
				begin
				cmd_state <= 1;
				cmd_cs <= 1'b0;				// turn on M25
				cmd_reg <= C0n1;				// a one or two byte command
				write_cmd_reg <= 1'b1;
				end
			end
		1: cmd_state <= 2;
		2: begin
			write_cmd_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin
				if (one_byte_cmd)
					begin
					cmd_cs <= 1'b1;	// turn off M25 if one byte command
					cmd_state <= 0;
					end
				else						// continue if reading status reg
					begin
					cmd_reg <= 8'h02;	// dummy
					cmd_state <= 3;
					write_cmd_reg <= 1'b1;
					end
				end
			end
		3: cmd_state <= 4;
		4: begin
			write_cmd_reg <= 1'b0;			// stop writing
			if (spi_done)
				begin		
				cmd_cs <= 1'b1;	// turn off M25 if reading status
				cmd_state <= 0;				
				end
			end
		default: ;
		endcase
		end
	end

/* Synchronize Apple slot clocks to CB clock		*/
	reg rw_1 = 0;
	reg reset_1 = 0;
	reg phase0_1 = 0;
	reg Q3_1;
	reg devsel_1 = 0;
	reg iosel_1 = 0;
	reg iostrobe_1 = 0;
//	reg m2b0_1 = 0;
	reg ph1_1 = 0;
//	reg user1_1 = 0;
	
	always @(posedge clk50) 	
		begin
		rw_1 <= BUS_R_W;
		rw_s <= rw_1;
      Q3_1 <= BUS_Q3; 
      Q3_s <= Q3_1;
		phase0_1 <= BUS_PHASE0;
		phase0_s <= phase0_1;
      ph1_1 <= BUS_PHASE1;
      ph1_s <= ph1_1;
		reset_1 <= BUS_RESET;
		reset_s <= reset_1;
		devsel_1 <= N_BUS_DEV_SEL;
		devsel_s <= devsel_1;
		iosel_1 <=	N_BUS_IO_SEL;
		iosel_s <= iosel_1;
		iostrobe_1 <= BUS_IOSTROBE;
		iostrobe_s <= iostrobe_1;
		bus_a_1 <= BUS_A;
		bus_a_s <= bus_a_1;
		bus_d_1 <= BUS_D;
		bus_d_s <= bus_d_1;
		end

`ifdef bbug
	assign bugin = 1'b0;		// replace 1'b0 with a wire/reg to send state to bugborder
`endif 
endmodule
















