`timescale 1ns / 1fs
//`define bbug 1
`define tester 1
`define clocked 1						// makes load_aux_mem & auxmem clocked by the 50MHz clock 
//`define vid_sync 1					// sync cb video to apple video
`default_nettype none

/* for //e & 12 bit color & SHR*/
module vga_video (
		  input clk50, 
		  input [7:0] char_data, 	// from font
		  input [7:0] ram_data,  	// from video data SRAM thru cb1.v
		  input [15:0] bus_addr,	// sync'd in cb1.v
		  input [7:0] bus_data,  	// sync'd in cb1.v
		  input ph1_s,					// sync'd in cb1.v
		  input phase0_s,	
		  input Q3,						// sync'd in cb1.v
		  input rw,
		  input reset_n,
		  input iosel_s,
		  input iostrobe_s,
		  input devsel_s,							
		  input BUS_SYNC,				// added 4/25/13 - signal only on slot 7
`ifdef bbug
			input bugin,				// use this to pass a signal for bugborder
`endif	
	//		input [15:0] debug,		// used to pass debug IO
			input init_done,			// if settings recieved from NVRAM
			input [7:0] txt_bk,		// text & background color settings from NVRAM
			input [3:0] bdr,			// border color setting from NVRAM 
			input bw_dhr,				// black & white dhr setting from NVRAM
	//		input opt12,				// optimize for 12 bit color setting from NVRAM // no longer used 5/19/13
			input [7:0] scb_data,		// input scanline from BRAM
	//		output reg [7:0] scb_addr,		// output scanline address in BRAM
			output [7:0] scb_addr,		// output scanline address in BRAM
			input [11:0] pal_data,		// input palette color from BRAM
			output [7:0] pal_addr,		// output palette color address in BRAM
			output update_time,

`ifdef clocked
			output reg auxmem,
`else
			output auxmem,
`endif
			output video_read,
			output v_sync,
			output h_sync,
			output reg [10:0] rom_addr,  
			output [16:0] ram_address, 
		  output [13:0] IDE_D,		// added 12/17/11 for 12 bit RGB
		  output clk25,				
		  output [5:0]	RGB,			// 6 bit RGB
		  output reg io_read,		// 4/16/13
		  output reg [7:0] io_reg	// 4/16/13
		  );

	reg rw_s;
	reg rw_1;
	reg reset_s;
	reg reset_1;

	reg [9:0]    h_count = 0;	// horz dot count 
   reg [9:0] 	 v_count = 47;	// vert line count
   reg	 clk1;	
   reg 	TEXT = 1;
   reg 	MIXED = 1;
   reg 	PAGE2 = 0;
   reg 	mode_80COL = 0;     
	reg   internal_slot3 = 1;
	reg WRITE_AUX = 0;      // writes to aux memory
	reg LINEAR_SHR = 0;		// $C029 bit 6
 	reg SHR = 0;						
	reg DHR = 0;
	reg monocolor = 0;
	reg [3:0] 	BORDER_COLOR_NUM = 4'b0000;  
	reg [3:0]	TEXT_COLOR_NUM = 4'b1111;	
	reg [3:0]	BACK_COLOR_NUM = 4'b0000;
	reg border_flag = 0;
	reg tb_flag = 0;
	
/* ----------------- bugborder uses VGA border color as data tester	-----------	*/	
`ifdef bbug
	reg bugborder = 0;		// added 9/30/10 for testing wires & single bit regs
	reg[1:0] bb_enable = 0;	// 0 = disable
	reg[7:0] bb_loc = 0;		// location to test $C0xx
	reg[7:0] bb_loc2 = 0;   // 2nd location to test $C0xx
	reg[7:0] bb_bit = 0;		// bit to test in returned value low nib for 1st location, high nib for 2nd
	reg[7:0] bb_io = 0;		// value returned
	reg[7:0] bb_io2 = 0;    // value returned for 2nd location
	reg bb_stat = 0;			// 0 = location 1, 1 = location 2
	reg bbrw = 0;				// mirrors the apple read/write line (rw) at the time of an I/O access
	reg bb_hit = 0;			// 1 = test address accessed
`endif
/* -----------------------------------------------------------------------------	*/
`ifdef tester
	reg[7:0] test42 = 0;		// input for various tests
`endif
	reg [7:0] 	 shifter7;
   wire [2:0] 	 scanline = v_count[3:1];
   reg [2:0] 	    count7;
   reg [6:0] 	    hc = 0;	
   reg [4:0] 	    vc = 0;

	reg [4:0] 	count28 = 0; 				// added for DHR
	reg [7:0]	shr_bytes = 0; 	 	// super hires bytes in 1 scanline	
   reg [15:0] shr_addr = 16'h2000;	// address of bytes for super hires	- 1
	wire disp40_active;
	wire shift_bitsum = disp40_active ? shifter7 > 0 : 1'b0;
   wire 	    GR =  ~(TEXT | ( MIXED &  vc[4] & vc[2])); // not text or mixed and >= line 20 
   reg 		 GR_r, GR_r2;
	wire	hgr;
   reg 		 hor_active_reg;			// gets set to hor_active (line being written) when count7 == 0
   reg [3:0] 	low_color_nibble;		// used to select lores color (16 colors)
   wire [11:0] lowres_color_val12;		// actual 12 bit color
   wire [5:0] 	lowres_color_val6;		// actual 6 bit color
	

	/* ------ handle 6 bit dhr, shr, border, text & background colors ----------  	*/
	reg [3:0] border_number;			// used to select border color (16 colors)
	wire [5:0] border_color6;			// actual 6 bit color (border)
	wire [11:0] border_color12;			// actual 12 bit color (border)
	reg [3:0] text_number;			// used to select text color (16 colors)
	wire [5:0] text_color6;			// actual 6 bit color (text)
	wire [11:0] text_color12;			// actual 12 bit color (text)
	reg [3:0] back_number;			// used to select background color (16 colors)
	wire [5:0] back_color6;			// actual 6 bit color (background)
	wire [11:0] back_color12;			// actual 12 bit color (background)
//	reg [3:0] shr_number;		  // used to select shr color (6 bit)
//	wire [5:0] shr_color6;
	/* -------------------------------------------------------------------------------	*/
	
	reg 	HIRES = 0;
	wire	dhires;							// double hires
	wire shires;							// super hires
	reg [5:0] color6;
	reg [11:0] color12;					// hires 12 bit color
	wire [5:0] dhr_color6;
	wire [11:0]	dhr_color12;			// double hires color 12 bit
	reg [3:0] dhrcolor_number;			// used to select double hires color (16 colors) 
	reg [11:0] grayscale;
	wire [5:0] grayscale6 = {grayscale[11:10], grayscale[7:6], grayscale[3:2]};
   reg 	hor_del2;						// gets set to hor_active_reg when count7 == 0
   reg [4:0] flash_count;   
   reg 		 flash, inverse;
   wire 	    flashing = flash ? flash_count[4] : 1'b0;      
	
	wire Q3_s = Q3;
									
   reg 	vid_del;
	reg  	AltCharSet,  mode_80STORE;	
	assign clk25 = clk1;	
	assign dhires = DHR & mode_80COL & GR;	
	assign 	shires = SHR & LINEAR_SHR;		// super hires	
   wire 	    h_blank = ( h_count > 639); 	// horz blanking - was 640 11/27/12
   wire 	    v_blank = ( v_count > 479);	// vertical blanking	- was 481 11/27/12

	
	wire vp = (h_blank | v_blank) ? 0 : (shifter7[0] ^ (inverse ^ flashing)); 
/* the following h_sync & v_sync is modified from original to	*/
/* be more compatible with HDTVs											*/
	assign 	    h_sync = (h_count < 659) || (h_count > 755);
	assign 	    v_sync = (v_count < 493) || (v_count > 494);
	
//	wire hor_active = (SHR == 1) ? (h_count >= 0) & (h_count < 642) : ~mode_80COL ? (h_count > (39-h_adj)) & (h_count < (612-h_adj)) : (h_count > (48-h_adj)) & (h_count < (618-h_adj));		
	
	wire hor_active = (SHR == 1) ? (h_count >= 0) & (h_count < 642) : ~mode_80COL ? (h_count > 39) & (h_count < 613) : (h_count > 48) & (h_count < 613);		
	wire   vert_active =  (SHR == 1) ? (v_count > 47) & ( v_count < 448) : (v_count > 47) & ( v_count < 432); 

	assign disp40_active =  hor_active_reg & hor_del2 & vert_active;

   wire 	    io_access = (bus_addr[15:8] == 8'b11000000);  //$C0xx area being accessed
	wire 	    video_bit = shifter7[0];
	wire allow_shift7 = (mode_80STORE & mode_80COL | DHR) ? 1'b1: h_count[0];  
`ifdef clocked
	reg load_aux_mem = 0;
`else
   wire load_aux_mem;
`endif
	reg 	aux_byte;

	assign 	update_time = (hgr & ~DHR) ?  ((count7 == 0) & (h_count[0] == 1'b1)) : DHR ? (h_count[1] == 1'b1) : (h_count[1] == 1'b0) ; 
	assign 	video_read = 1'b1;  // SRAM datasheet says /OE can be low for both read and write 

/* --- RGB 6 bit ------------------------------------------------------------------------------------------	*/
//	wire [5:0] GR_RGB =  dhires ? dhr_color6 : (HIRES == 0) ? lowres_color_val6 : video_bit ? color6 : 6'b0; // was just video_bit for color

	wire [5:0] GR_RGB = (dhires && monocolor && HIRES) ? grayscale6 : (dhires && HIRES) ? dhr_color6 : (HIRES == 0) ? lowres_color_val6 : color6; 
/*
	assign	RGB = (h_blank | v_blank)  ? 6'b0 : (~disp40_active) ? border_color6 : (SHR) ? shr_color6 :
		    (GR_r2) ? GR_RGB : (({vp,vp,vp,vp,vp,vp} & text_color6[5:0]) ^ ({~vp,~vp,~vp,~vp,~vp,~vp} & back_color6[5:0])) ;
*/
	assign RGB = SHR ? {(IDE_D[11] & IDE_D[10]), (IDE_D[9] & IDE_D[8]), (IDE_D[7] & IDE_D[6]), (IDE_D[5] & IDE_D[4]), 
						(IDE_D[3] & IDE_D[2]), (IDE_D[1] & IDE_D[0])} : (h_blank | v_blank)  ? 6'b0 : (~disp40_active) ? border_color6 :/* (SHR) ? shr_color6 : */
		    (GR_r2) ? GR_RGB : (({vp,vp,vp,vp,vp,vp} & text_color6[5:0]) ^ ({~vp,~vp,~vp,~vp,~vp,~vp} & back_color6[5:0]));
	colors_6 colors_6 (
		.border_number(border_number),	// input
		.text_number(text_number),			// input 
		.back_number(back_number),			// input
		.dhrcolor_number(dhrcolor_number),	// input	
		.nibble(low_color_nibble),			// input
//		.shr_number(shr_number),			// input	
		.border_color6(border_color6),	// output
		.text_color6(text_color6),		// output
		.back_color6(back_color6),		// output
		.dhr_color6(dhr_color6),			// output
		.RGB_val(lowres_color_val6)     //,		// output
//		.shr_color6(shr_color6)
			);
			
/* --- RGB 12bit using IDE connector and resistors	-------------------------------------------------------	*/
	reg[11:0] shr_color12 = 0;	// 12 bit
	wire [11:0] text12 = (({vp,vp,vp,vp,vp,vp,vp,vp,vp,vp,vp,vp} & text_color12[11:0]) ^ ({~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp,~vp} & back_color12[11:0]));
	//	wire [11:0] GR_RGB12 = dhires ? dhr_color12 : (HIRES == 0) ? lowres_color_val12 : /* video_bit ? */ color12; // : 12'b0; 
wire [11:0] GR_RGB12 = (dhires && monocolor && HIRES) ? grayscale : (dhires && HIRES) ? dhr_color12 : (HIRES == 0) ? lowres_color_val12 : /* (video_bit | shifter7[1]) ? */  color12; // : 12'b0; 

	assign IDE_D[13] = v_sync;
	assign IDE_D[12] = h_sync;
	assign IDE_D[11:0] = (h_blank | v_blank)  ? 12'b0 : (~disp40_active) ? border_color12 : (SHR) ? shr_color12 :
							(GR_r2) ? GR_RGB12 : text12 ;
							
	colors_12 colors_12 (
		.border_number(border_number),	// input 
		.text_number(text_number),			// input  
		.back_number(back_number),			// input
		.dhrcolor_number(dhrcolor_number),	// input	
		.nibble(low_color_nibble),			// input		
		.border_color12(border_color12),	// output
		.text_color12(text_color12),		// output
		.back_color12(back_color12),		// output
		.dhr_color12(dhr_color12),			// output
		.lowres_color_val12(lowres_color_val12)		// output
			);
`ifdef vid_sync
/* --- sync CB video with Apple video ---	*/
	reg [8:0] synccnt = 0;
	reg acb_sync = 0;
	reg [1:0]sync_state = 0;
	always @ (posedge clk50) //(posedge phase0_s) // clk50)
		begin
		if(reset_n == 1'b0)
			begin
			synccnt <= 0;
			acb_sync <= 1'b0;
			sync_state <= 0;
			end
		else
		case (sync_state)
			0: begin
				acb_sync <= 1'b0;	
				if(~BUS_SYNC) sync_state <= 1; // apple's sync pulse goes low
				else sync_state <= 0;
				end
			1: begin
				synccnt <= synccnt + 1;					// time the length of the sync pulse
				if ((synccnt < 450) && (BUS_SYNC)) 				//  not long enough to be v_sync *** 55 max?
					begin
					synccnt <= 0;
					sync_state <= 0;
					end
		//		if (synccnt == 255)  // vert sync apple // (synccnt == 2356)  end of vertical blanking on Apple
				if (synccnt == 450)
					begin
					sync_state <= 2;
					acb_sync <= 1'b1;					
					end
				end
			2: begin
				acb_sync <= 1'b0;						
				if(BUS_SYNC)
					begin
					synccnt <= 0;
					sync_state <= 0;
					end
				end
			default: sync_state <= 0;
		endcase
		end
`endif
/* --------- increments through the dots -------	*/
	reg line_dbl = 0;
	
   always @(posedge clk50) 
		begin
		clk1 <= ~clk1;

`ifdef vid_sync
		else if(acb_sync ) // & (h_count == 0)/*& (v_count == 523)*/)
			begin
			v_count <= 448 + test42;
//			clk1 <= 0;
			end
`endif
	//	else // not reset
			begin
			if(clk1) 													
				begin				
				if(disp40_active)
					begin										// SHR address calc & line doubler
					if (h_count[1:0] == 2'b10)			
						begin
						shr_addr <= shr_addr + 1;
						if(shr_addr == 16'hA000) shr_addr <= 16'h2000;
						shr_bytes <= shr_bytes + 1;								
						if (shr_bytes == 8'h9F) 
							begin
							if (line_dbl == 0) shr_addr <= shr_addr - shr_bytes; // line doubler
							shr_bytes <= 8'h00;
							line_dbl <= ~line_dbl; // flip doubler flag
							end
						end
					end	// end if(disp40_active)
				if ((h_count < 6 + test42)  || (count28 == 27)) count28 <= 0;
//				if ((h_count < 6 + test42)  || (count28 == 27) || (h_count > 612 )) count28 <= 0;
				else count28 <= count28 + 1;

				if (h_count == 799) 							// end of horizontal count 0-799  
					begin
					h_count <= 0; 								// start over
					count7 <= 0;  								// reset 0-6 counter
			//		count28 <= 0;
					shr_bytes <= 8'h00;
					if ((v_count == 447) || (v_count < 48)) 
						begin
						shr_addr <= 16'h2000;
						line_dbl <= 1'b0;
						end
					if(v_count == 524) v_count <= 0;		// ( last line 0-524 start over)
					else v_count <= v_count +1;			// not last line so start next line
					end  						
				else 												// not end of horizontal count 0-799
					begin
					h_count <= h_count + 1;					// increment the horizontal count 0-799
					if(allow_shift7) 				// every clk1 for 80c, every other clk1 for 40c
						begin
						if(count7 == 6) count7 <= 0;	// end of 0 to 6 count so reset 
						else count7 <= count7 + 1;		// not end so increment	
						end    
					end
				end // clk1
			else // not clk1 
				begin // not clk1
				if(count7 == 3'b0) rom_addr <= {ram_data[7:0], scanline};		// changed to non-blocking assign																									
				end // else: !if(clk1)
			end // not reset
		end // posedge clk50
		
/* --------------- end increments through the dots ------ 	*/  

	wire X_PAGE2 = PAGE2 & ~mode_80STORE; 	
   reg 	high_bit;
`ifdef clocked
	always @(posedge clk50)
		begin
		load_aux_mem <= (mode_80COL && (TEXT | ( MIXED & vc[4] & vc[2]) | DHR) && aux_byte) | SHR;
		auxmem <= (mode_80STORE & PAGE2) | WRITE_AUX;  
		end
`else
   assign load_aux_mem = (mode_80COL & ((~GR_r2 | DHR) & aux_byte)) | SHR;
	assign auxmem <= (mode_80STORE & PAGE2); 
`endif
	wire [5:0] high_vid_addr = (hgr) ? {1'b0, X_PAGE2, ~X_PAGE2, scanline} : {4'b0000, X_PAGE2, ~X_PAGE2};	// line
	assign hgr = (HIRES & GR_r2);						  
	wire [3:0] addr_sum = {~hc[5],~hc[5],hc[4],hc[3]} +  {vc[4],vc[3],vc[4],vc[3]} + 1;		
	wire [15:0] apple_video_addr = (SHR) ? shr_addr : {high_vid_addr, vc[2:0], addr_sum ,hc[2:0]};
   reg [15:0]  apple_video_addr_reg;
	
	assign      ram_address = {load_aux_mem, apple_video_addr_reg};	// this is the one used for getting ram_data from buffer 
	reg hor_del1 = 0;

	always @(posedge clk50)
		begin
		if (~disp40_active & clk1) 
			begin
`ifdef bbug				
			border_number <= border_flag ? BORDER_COLOR_NUM + bugborder : init_done ? bdr + bugborder : BORDER_COLOR_NUM;
`else				
			border_number <= border_flag ? BORDER_COLOR_NUM : init_done ? bdr : BORDER_COLOR_NUM;
`endif
			text_number   <= tb_flag ? TEXT_COLOR_NUM : init_done ? txt_bk[7:4] : TEXT_COLOR_NUM;		
			back_number	  <= tb_flag ? BACK_COLOR_NUM : init_done ? txt_bk[3:0] : BACK_COLOR_NUM; 		
			end
		
		if (SHR  || LINEAR_SHR && ~TEXT) 
				begin
				hor_active_reg <= hor_active;
				hor_del2 <= hor_active_reg;
				apple_video_addr_reg <= apple_video_addr;
				end			
		else 
      if(clk1 && allow_shift7 && ~hgr && ~SHR) 					// every clk1 for 80c, every other clk1 for 40c
			begin						
			if (count7 == 0) 
				begin
				hor_active_reg <= hor_active;
				hor_del2 <= hor_active_reg; 			// was hor_del2 <= hor_active_reg;
				apple_video_addr_reg <= apple_video_addr;				
				end
			end 
		else	
      if(clk1 && hgr && ~SHR) 				
			begin						
			if ((count7 == 0) && ~mode_80COL && h_count[0])   
				begin
				hor_active_reg <= hor_active;
				hor_del2 <= hor_active_reg; 			// was hor_del2 <= hor_active_reg;
				apple_video_addr_reg <= apple_video_addr;				
				end
			else
			if ((count7 == 0) && mode_80COL)
				begin
				hor_active_reg <= hor_active;
				hor_del1 <= hor_active_reg; 			// was hor_del2 <= hor_active_reg;
				apple_video_addr_reg <= apple_video_addr;	
				end
			if (mode_80COL) hor_del2 <= hor_del1;			
			end 
   end
		
	reg [6:0] aux_data0 = 0;
	reg [6:0] main_data1 = 0;
	reg [6:0] aux_data2 = 0;
	reg [6:0] main_data3 = 0;

/* ------------ Super Hires 12 bit ------------------- */	
	reg [7:0] shr_data;
	wire shr_resmode = scb_data[7]; 		// 640 if = 1

	wire fill = (shr_resmode == 0) & (scb_data[5] == 1) & (((h_count[1] == 1'b0) && (shr_data[7:4] == 0)) || ((h_count[1] == 1'b1) && (shr_data[3:0] == 0)));
	wire [3:0] palette_addr = scb_data[3:0];
	wire [8:0] vfix = v_count - 48;  // was 47 but 48 fixes some of palette glitches 
	wire [7:0] a2line = vfix[8:1]; // vfix/2

	always @(posedge clk50)
		begin
	
		if(SHR & clk1)
			begin	
			if(h_count[1:0] == 2'b11) shr_data <= ram_data[7:0];
			if (~fill) shr_color12 <= pal_data;
			end
		end
	assign scb_addr = a2line;
	assign pal_addr = (shr_resmode == 1) ? ((h_count[1:0] == 2'b00) ? {palette_addr[3:0], 2'b10, shr_data[7:6]}:
							(h_count[1:0] == 2'b01) ? {palette_addr[3:0], 2'b11, shr_data[5:4]} :
							(h_count[1:0] == 2'b10) ? {palette_addr[3:0], 2'b00, shr_data[3:2]} :
							{palette_addr[3:0], 2'b01,shr_data[1:0]}):
							((h_count[1] == 1'b0) ? {palette_addr[3:0], shr_data[7:4]} : {palette_addr[3:0], shr_data[3:0]});

/* --------------------------- DHR --------------------------------------	*/

	wire ard0 = (count28 == 27); //(count28 == 27);
	wire mrd1 = (count28 == 3); //(count28 == 2);
	wire ard2 = (count28 == 11); // (count28 == 11);
	wire mrd3 = (count28 == 19); //(count28 == 19);

	wire[3:0] l_color = ~disp40_active ? border_number[3:0] : main_data3[6:3] ;

/* DHR */
	always @(posedge clk50)
		begin
		if (DHR & clk1 & allow_shift7) // (dhires)
			begin
			if (ard0) aux_data0 <= ram_data[6:0];
			if (mrd1) main_data1 <= ram_data[6:0];
			if (ard2) aux_data2 <= ram_data[6:0];
			if (mrd3) main_data3 <= ram_data[6:0];
			end
		end 
/* DHR color per pixel */
reg [3:0] c_number = 0;

	always @(posedge clk50)
		begin
			if (clk1 & DHR & allow_shift7)
			begin
			if ((h_count < 614) && DHR) 
				begin
				case (count28[4:2]) // 
				3'b000:	c_number <= aux_data0[3:0];
				3'b001:	c_number <= { main_data1[0], aux_data0[6:4] };
				3'b010:	c_number <= main_data1[4:1];  // >> 1
				3'b011:	c_number <= { aux_data2[1:0], main_data1[6:5] };
				3'b100:	c_number <= aux_data2[5:2];
				3'b101:	c_number <= { main_data3[2:0], aux_data2[6] };
				3'b110:	c_number <= main_data3[6:3] ;
				3'b111: 	c_number <= 4'b0000;	// should never happen
				default: c_number <= 4'b0000;
				endcase
				grayscale <= {c_number,c_number,c_number};
				dhrcolor_number <= c_number;
				end
			end
		end
/* --------------------------------------------------------------	*/
	always @(posedge clk50 )
		begin		
		if(clk1 && h_count[0] && hgr &&  ~DHR  && ~SHR )	// was ~dhires
			begin
			vid_del <= video_bit;
			if (count7 == 6)			
				begin
				shifter7 <= ram_data[6:0];
				high_bit <= ram_data[7];
				end
			else 
				begin
				shifter7 <= {1'b0, shifter7[7:1]};  // right shift 1 bit
				end
			end
		
		else if(clk1 && allow_shift7 && ~hgr  && ~SHR)  
				begin
					if(count7 == 0) 
						begin
						shifter7 <= (~ram_data[7]) ? char_data [7:0] : ~char_data[7:0];
						low_color_nibble <= v_count[3] ? ram_data[7:4] : ram_data[3:0];				
						flash  <= AltCharSet ? 0 : (ram_data[6] & ~ram_data[7]);		
						inverse <=  ~ram_data[7];	
						end
					else	// not count7 == 0 						
						begin
						shifter7 <= {1'b0, shifter7[7:1]};  // right shift 1 bit
						end // else not count7 == 0	 
				end
		end // always @(posedge clk50)

	wire spec  = allow_shift7 & (count7 == 0) & (hc == 64);
	wire next_vc = clk1 & spec & ((aux_byte == 0) | ~mode_80COL);

`define WHITE  6'b111111
`define BLACK  6'b000000
`define GREEN  6'b001000
`define PURPLE 6'b110010
`define RED    6'b110100
`define BLUE   6'b001011  

`define DEEP_RED 		6'b110000
`define BROWN			6'b100100
`define ORANGE			6'b110100
`define DARK_GREEN	6'b000100
`define DARK_GRAY		6'b010101
`define YELLOW			6'b111100
`define DARK_BLUE		6'b000010	// same as lores deep_blue 
`define LIGHT_GRAY	6'b101010
`define PINK			6'b111010
`define MEDIUM_BLUE	6'b000011
`define LIGHT_BLUE	6'b011011
`define AQUAMARINE	6'b001110


	wire       odd_dot = ~h_count[1];

/* hires colors per pixel */ 
	reg [11:0] last12 = 12'h000;
	reg [5:0] last6 = 6'b0;
	reg non_color = 0;
always @ (posedge clk1)
	begin
	last12 <= color12;
	last6 <= color6;
	if (video_bit && vid_del) 
			begin
			color12 <=12'hFFF;	// 12 bit
			color6 <= 6'b111111; // 6 bit
			if (shift_bitsum != 1) non_color <= 1'b1;
			end
	else 	
	if(video_bit && ~vid_del)
		begin
		if (non_color)
			begin 
			color12 <= last12;
			color6 <= last6;
			non_color <= 1'b0;
			end
		else	
		if(odd_dot) 
			begin  
			if(high_bit) 
				begin
				color12 <= 12'h6AF; //12'h22F;
				color6 <= `BLUE;
				end
			else 
				begin
				color12 <= 12'hD2D;
				color6 <= `PURPLE;
				end
			end 
		else
			begin
			if(high_bit) 
				begin
				color12 <= 12'hF60; // color12 <= 12'hF98;
				color6 <= `RED;
				end
			else 
				begin
				color12 <= 12'h1D0;
				color6 <= `GREEN;
				end
			end
		end
	else 
	if (~video_bit && ~vid_del) 
		begin
		color12 <=12'h000;	// 12 bit 
		color6 <= 6'b000000; // 6 bit
		non_color <= 1'b1;
		end
	end	

	always @(posedge clk50)
		begin
		if(clk1 && allow_shift7) 	// every clk1 for 80c, every other clk1 for 40c
			begin
			if(~(hor_active && vert_active)) // h & v_counts = horz not 40-607, vert 48-431
				begin // not counting when display screen not active (we're in blanking)
				hc <= 7'b0011000;	// ($18 - 24 ) changed to 7 bits 11/10/09
				aux_byte <= GR ? 0 : 1;			// was 1 before aux_byte test
				end // if(~(hor_active && vert_active))    
			else 																		// (hor_active && vert_active)
				begin 															// video is active
				if(count7 == 0)
					begin
					aux_byte <= ~aux_byte; 
					GR_r <= GR;
					GR_r2 <= GR_r;					
					if((aux_byte == 1) || ~mode_80COL)  // always if not 80c othewise every other time
						begin												
						hc <= hc +1;								// increment counter 24 to 63
						end // if((aux_byte == 1) || ~mode_80COL)
					end //if(count7 == 0) 
				end // else: !if(~(hor_active && vert_active))
			end // if (clk1  &&  allow_shift7)
		end // always @ (posedge clk50)

   always @(posedge clk50) 
		begin
		if (~vert_active) vc <= 5'b00000;	
		else // vert_active
			begin // vert_active
			if(next_vc) 
				begin 
				if((v_count == 48) || (vc == 24)) 	
					begin
					flash_count <= flash_count +1; // changed to non-blocking assign;
					vc <= 5'b00000; 	
					end   
				else  
				if (v_count[3:0] == 4'b1111)  vc <= vc + 1;  // increment char line every 16 scan lines
			end // next_vc
		end // vert_active
		end // poseedge clk50

   // decoding Apple soft switches 
	always @(negedge Q3_s)
		begin 
		if(ph1_s == 0)
			begin
			io_read <= 0;
			if(reset_s == 1'b0) 
				begin

`ifdef tester
				test42 <= 8'b00000000;
`endif
				TEXT <= 1;
				PAGE2 <= 0;
				HIRES <= 0;
				MIXED <= 0;  
				AltCharSet <= 0;
				mode_80COL <= 0;
				mode_80STORE <= 0;
				DHR <= 0;
				SHR <= 0;
				WRITE_AUX <= 0;
				io_read <= 0;

				/* ------------------------------------------------------------------- */
				end // if(reset_s == 0) 
			else // not reset
			if(io_access && (rw_s | ~rw_s)) 
				begin 
`ifdef bbug
				if((bb_enable[1:0] == 2'b01)  && (bb_loc[7:0] == bus_addr[7:0])) 
					begin
					bb_io[7:0] <= bus_data[7:0];
					bb_stat <= 1'b0;
					bbrw <= rw_s;	// read/write status write = 0, read = 1
					bb_hit <= 1'b1;
					end
				else 
				if((bb_enable[1:0] == 2'b10) && ((bb_loc[7:0] == bus_addr[7:0]) || (bb_loc2[7:0] == bus_addr[7:0])))
					begin
					bb_hit <= 1'b1;
					if (bb_loc[7:0] == bus_addr[7:0]) 
						begin
						bb_io[7:0] <= bus_data[7:0];
						bb_stat <= 1'b0;
						bbrw <= rw_s;	// read/write status write = 0, read = 1
						end
					if (bb_loc2[7:0] == bus_addr[7:0]) 
						begin
						bb_io2[7:0] <= bus_data[7:0];
						bb_stat <= 1'b1;
						bbrw <= rw_s;	// read/write status write = 0, read = 1
						end
					end
				else bb_hit <= 1'b0;
`endif
					begin
					case(bus_addr[7:0])
					8'h50:      TEXT <= 0;
					8'h51:      TEXT <= 1;
					8'h52:      MIXED <= 0;
					8'h53:      MIXED <= 1;
					8'h54:      PAGE2 <= 0;
					8'h55:      PAGE2 <= 1;   
					8'h56:      HIRES <= 0;
					8'h57:      HIRES <= 1;				  
					8'h5E:		  DHR <= 1;		// double hires on 11/7/09
					8'h5F:		  DHR <= 0;		// double hires off	
					endcase // case(bus_addr[7:0])
					
					if(rw_s == 1'b0)
						begin           // writes that switch
						io_read <= 0;
						case(bus_addr[7:0])
						
						8'h00: mode_80STORE <= 0; //$C000
						8'h01: mode_80STORE <= 1;

						8'h04:  WRITE_AUX <= 0;
						8'h05:  WRITE_AUX <= 1;
						8'h0C:  mode_80COL <= 0;
						8'h0D:  mode_80COL <= 1;
						
						8'h0E:  AltCharSet <= 0;
						8'h0F:  AltCharSet <= 1; //$C00F

		//				8'h21: monocolor <= bus_data[7];		// 1 = color disabled for DHR
						8'h29: 
							begin 
							SHR <= bus_data[7];
							LINEAR_SHR <= bus_data[6];
							monocolor <= bus_data[5];			// does the same as $C021 bit 7
							end
						
						/* --- looks for border, text & background color changes --- */
						8'h34: 	// $C034 lower 4 bits is color#
							begin
							border_flag <= 1'b1;
							BORDER_COLOR_NUM <= bus_data[3:0];
							end
`ifdef tester
						8'h42: 
							begin
							test42 <= bus_data[7:0];	// input used for various tests
							end
`endif
`ifdef bbug
						8'h69: bb_enable <= bus_data[1:0]; // pseudo I/O for bugborder input 1 = on
						8'h6A: bb_loc <= bus_data[7:0];  // pseudo I/O for bugborder I/O location to monitor
						8'h6B: bb_loc2 <= bus_data[7:0];	// pseudo I/O for bugborder 2nd I/O loc. to monitor
						8'h6C: bb_bit <= bus_data[7:0];	// pseudo I/O for bugborder bit to test
`endif
						8'h22:	// $C022 high 4 bits = text color# low 4 bits = background color#
							begin
							tb_flag <= 1'b1;
							TEXT_COLOR_NUM <= bus_data[7:4];
							BACK_COLOR_NUM <= bus_data[3:0];
							end			
							
/* ------------------------------------------------------------- */
						endcase
						end // if(~rw_s) 
					if(rw_s == 1'b1)	// read a switch
						begin           // writes that switch
						case(bus_addr[7:0])				
						/* --- registers  border, text & background color changes --- */

						8'h22:	// $C022 high 4 bits = text color# low 4 bits = background color#
							begin
							io_read <= 1;
							io_reg [7:0] <= ~tb_flag ? txt_bk : {TEXT_COLOR_NUM, BACK_COLOR_NUM};
							end
						8'h29: 
							begin 
							io_read <= 1;
							io_reg [7:0] <= {SHR,LINEAR_SHR,monocolor,5'b0};
							end
						8'h34: 	// $C034 lower 4 bits is color#
							begin
							io_read <= 1;
							io_reg [3:0] <= ~border_flag ? bdr : {4'bz, BORDER_COLOR_NUM};
							end
						default: io_read <= 0;
						endcase
						end		
					end 
				end // if io_access  
			end // if ph1 == 0
		end // always @ (negedge Q3)

/* sync apple clocks with CB */

   always @(posedge clk50) 
		begin
		rw_1 <= rw;
		rw_s <= rw_1;
		reset_1 <= reset_n;		// this probably shouldn't be done
		reset_s <= reset_1;		// it's sync'd in cb1.v
		end
`ifdef bbug

/* changes border color to indicate reg or wire value 9/30/10 */
/* bugborder is added to the border_color_num	*/

   always @(posedge clk50) 
		begin
		if(bb_enable[1:0] == 2'b01) 
			begin
			if(bb_hit)
				begin
				case (bb_bit[3:0])
				8'h00: bugborder <= bb_io[0];
				8'h01: bugborder <= bb_io[1];
				8'h02: bugborder <= bb_io[2];
				8'h03: bugborder <= bb_io[3];
				8'h04: bugborder <= bb_io[4];
				8'h05: bugborder <= bb_io[5];
				8'h06: bugborder <= bb_io[6];
				8'h07: bugborder <= bb_io[7];
				8'h08: bugborder <= 1'b1;
				default: bugborder <= 0;
				endcase
				end
			end
		else 
		if(bb_enable[1:0] == 2'b10)
			begin
			if ((bb_stat == 1'b0) && bb_hit)
				begin
				case (bb_bit[3:0])
				8'h00: bugborder <= bb_io[0];
				8'h01: bugborder <= bb_io[1];
				8'h02: bugborder <= bb_io[2];
				8'h03: bugborder <= bb_io[3];
				8'h04: bugborder <= bb_io[4];
				8'h05: bugborder <= bb_io[5];
				8'h06: bugborder <= bb_io[6];
				8'h07: bugborder <= bb_io[7];
				8'h08: bugborder <= bb_stat;	// location 1 accessed
				default: bugborder <= 1'b0;		// off state if 8'h08
				endcase
				end
			else 
			if ((bb_stat == 1'b1) && bb_hit)
				begin
				case (bb_bit[7:4])
				8'h00: bugborder <= bb_io2[0];
				8'h01: bugborder <= bb_io2[1];
				8'h02: bugborder <= bb_io2[2];
				8'h03: bugborder <= bb_io2[3];
				8'h04: bugborder <= bb_io2[4];
			 	8'h05: bugborder <= bb_io2[5];
				8'h06: bugborder <= bb_io2[6];
				8'h07: bugborder <= bb_io2[7];
				8'h08: bugborder <= bb_stat;	// location 2 accessed
				default: bugborder <= 1'b1;
				endcase
				end
			end
		else
		
		bugborder <= 1'b0;			// any single bit reg or wire can be placed here

		end
`endif
endmodule 
/* --------------------------------------------------------------------	*/

